from quark.plugin_manager import factory

from quark_plugin_tsp.tsp_graph_provider import TspGraphProvider
from quark_plugin_tsp.tsp_qubo_mapping_dnx import TspQuboMappingDnx
from quark_plugin_tsp.classical_tsp_solver import ClassicalTspSolver

def register() -> None:
    factory.register("tsp_graph_provider", TspGraphProvider)
    factory.register("tsp_qubo_mapping_dnx", TspQuboMappingDnx)
    factory.register("classical_tsp_solver", ClassicalTspSolver)



import logging
import math
from dataclasses import dataclass
from typing import override

import networkx as nx
from quark.core import Core, Data, Result
from quark.interface_types import Graph, InterfaceType, Other


@dataclass
class TspGraphProvider(Core):
    """A module for creating random weighted complete graphs for the TSP problem.

    :param nodes: The number of nodes in the graph
    :param seed: The seed for the random number generator
    """

    nodes: int = 10
    seed: int = 42

    @override
    def preprocess(self, data: InterfaceType) -> Result:
        # Create a random graph without edges
        g = nx.random_geometric_graph(self.nodes, radius=0.0, seed=self.seed)
        pos = nx.get_node_attributes(g, "pos")

        # Add edges with weights equal to the Euclidean distance between the nodes
        for i in range(len(pos)):
            for j in range(i + 1, len(pos)):
                dist = math.hypot(pos[i][0] - pos[j][0], pos[i][1] - pos[j][1])
                g.add_edge(i, j, weight=dist)

        self._graph = g
        return Data(Graph.from_nx_graph(self._graph))

    @override
    def postprocess(self, data: InterfaceType) -> Result:
        match data:
            case None:
                logging.warn("No route found")
                return Data(None)
            case Other(list(l)):
                if any(node not in l for node in self._graph.nodes()):
                    logging.warn("Invalid route: Not all nodes were visited")
                    return Data(None)
                logging.info(f"All {len(self._graph.nodes())} nodes were visited")

                if len(l) != len(self._graph.nodes()):
                    logging.warn(
                        "Invalid route: Some nodes were visited more than once",
                    )
                    return Data(None)
                logging.info("All nodes were visited exactly once")

                distance = 0
                node_id = l[0]
                for next_node_id in l[1:]:
                    try:
                        edge = self._graph[node_id][next_node_id]
                    except KeyError:
                        logging.warn(f"Invalid route: Edge {node_id} -> {next_node_id} does not exist")
                        return Data(None)

                    distance += edge["weight"]
                    node_id = next_node_id

                logging.info("Route found!")
                logging.info(f"distance: {distance}")

                return Data(Other(distance))

            case _:
                # TODO add error message
                raise TypeError



import logging
from dataclasses import dataclass
from typing import override

import dwave_networkx as dnx
from quark.core import Core, Data, Result
from quark.interface_types import Graph, Other, Qubo


@dataclass
class TspQuboMappingDnx(Core):
    """A module for mapping a graph to a QUBO formalism for the TSP problem."""

    @override
    def preprocess(self, data: Graph) -> Result:
        self._graph = data.as_nx_graph()
        q = dnx.traveling_salesperson_qubo(self._graph)
        return Data(Qubo.from_dict(q))

    @override
    def postprocess(self, data: Other) -> Result:
        d = data.data
        # Only keep the tuples corresponding to |1> qubits
        tuples = (item[0] for item in d.items() if item[1])
        # Sort in order of time steps
        sorted_tuples = sorted(tuples, key=lambda x: x[1])

        path = [x[0] for x in sorted_tuples]
        time_steps = [x[1] for x in sorted_tuples]

        if time_steps != list(range(self._graph.number_of_nodes())):
            logging.warn("Invalid route")
            return Data(None)
        return Data(Other(path))



from dataclasses import dataclass
from typing import override

import networkx as nx
from quark.core import Core, Data, Result
from quark.interface_types import Graph, Other


@dataclass
class ClassicalTspSolver(Core):
    """Module for solving the TSP problem using a classical solver."""

    @override
    def preprocess(self, data: Graph) -> Result:
        self._solution = nx.approximation.traveling_salesman_problem(data.as_nx_graph(), cycle=False)
        return Data(None)

    @override
    def postprocess(self, data: None) -> Result:
        return Data(Other(self._solution))
